#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include <dlfcn.h>

#include "XMaidP.h"
#include "xmaidmod.h"

XMaidModule *
xmaidmodule_new(void)
{
	XMaidModule *p;
	if ((p = malloc(sizeof(*p))) == NULL) {
		perror("xmaidmodule_new");
		exit(1);
	}
	p->handler = NULL;
	p->info    = NULL;
	return p;
}

void
xmaidmodule_delete(XMaidModule *p)
{
	if (p) {
#ifndef NOMODULE
		if (p->handler)
			dlclose(p->handler);
#endif NOMODULE
		free(p);
	}
}

void
xmaidmoduledata_init(XMaidModuleData *p)
{
	p->data = NULL;
	p->anim = NULL;
	p->x = 0;
	p->y = 0;
	p->xbak = 0;
	p->ybak = 0;
	p->root_bak  = None;
	p->root_work = None;
	p->level = 0;
}

void
xmaidmoduledata_delete(XMaidModuleData *p)
{
	if (p){
		if (p->data)
			free(p->data);
		free(p);
	}
}

#define MODULE			(module)
#define MODULEINFO		(module->info)
#define ARGC			(xmw->application.argc)
#define ARGV			(xmw->application.argv)
#define ROOT			(xmw->xmaid.root)

#ifdef NOMODULE

extern XMaidModuleInfo focuswin_info;
#ifdef XMULTI
extern XMaidModuleInfo nadenade_info;
extern XMaidModuleInfo osouji_info;
#endif

typedef struct {
	char *name;
	XMaidModuleInfo *info;
} XMaidModuleInfoTable;

#define MODTABLE(name)  { #name, & name ## _info }

static XMaidModuleInfoTable module_table[] = {
	MODTABLE(focuswin),
#ifdef XMULTI
	MODTABLE(nadenade),
	MODTABLE(osouji),
#endif	
};

XMaidModule *
xmaid_module_open(XMaidWidget xmw, const char *name)
{
	XMaidModule *m = NULL;
	int i;
	for (i=0;i<sizeof(module_table)/sizeof(module_table[0]);i++)
		if (!strcmp(name, module_table[i].name))  {
			m = xmaidmodule_new();
			m->info = module_table[i].info;
			break;
		}	
	return m;
}

#else /* NOMODULE */

#ifndef RTLD_LAZY
#define RTLD_LAZY 1
#endif

static XMaidModule *
xmaidmodule_load(char *path, const char *name)
{
	XMaidModule *h = NULL;
	void *handler;
	char *buf;
	if ((buf = malloc(strlen(name) + sizeof("__info") + 1)) == NULL)
		return NULL;
	if ((handler = dlopen(path, RTLD_LAZY))) {
		h = xmaidmodule_new();
		h->handler = handler;
#ifdef linux
		sprintf(buf, "%s_info", name);
#else
		sprintf(buf, "_%s_info", name);
#endif
		h->info = dlsym(handler,buf);
	} else {
		fprintf(stderr, "xmaidmodule_load[%s:%s]: failed\n",
				path, name);
		exit(1);	
	}
	free(buf);

	if (!h->info) {
		fprintf(stderr, "module contain no infomation.\n");
		exit(1);
	}

#if 0
	fprintf(stderr, "module_load: %p %p %p\n", h->handler, h->data, h->info);
#endif

	return h;
}

XMaidModule *
xmaid_module_open(XMaidWidget xmw, const char *name)
{
	char *buf;
	XMaidModule *m = NULL;
	if ((buf = malloc(strlen(xmw->xmaid.modulepath) + strlen(name)
					  + sizeof("mod_.so") + 1))) {
		sprintf(buf, "%s/mod_%s.so", xmw->xmaid.modulepath, name);
		m = xmaidmodule_load(buf, name);
		free(buf);
	}
	return m;
}

#endif /* NOMODULE */

XMaidModuleData *
xmaid_moduledata_open(XMaidWidget xmw, XMaidModule *module, int num)
{
	int i;

	XMaidModuleData *p, *q;
	if ((p = malloc(sizeof(*p) * num)) == NULL) {
		perror("xmaidmoduledata_new");
		exit(1);
	}
	q = p;
	for (i=0;i<num;i++,p++) {
		xmaidmoduledata_init(p);
		if (MODULEINFO->datasize) {
			if ((p->data = malloc(MODULEINFO->datasize)) == NULL) {
				perror("xmaidmoduledata_open");
				exit(1);
			}
		}
	}
	return q;
}

void
xmaid_module_init(XMaidWidget xmw, XMaidModule *module)
{
	if (MODULEINFO->classinit)
		MODULEINFO->classinit();

	/* read MODULE's options and resources */
	if (MODULEINFO->options) {
		XrmDatabase db = XtScreenDatabase(XtScreen((Widget)xmw));
		XrmParseCommand(&db, 
						MODULEINFO->options, 
						MODULEINFO->noptions,
						XtName((Widget)xmw), &ARGC, ARGV);
	}	
}

void
xmaid_moduledata_init(XMaidWidget xmw, XMaidModule *module, 
					  XMaidModuleData *moddat)
{
	if (MODULEINFO->resources) {
		XtGetSubresources((Widget)xmw, moddat->data, 
						  MODULEINFO->name, "Module",
						  MODULEINFO->resources,
						  MODULEINFO->nresources, NULL, 0);
	}

	if (MODULEINFO->init)
		MODULEINFO->init(xmw, moddat);

	if (!ROOT && MODULEINFO->event)
		XtAddEventHandler((Widget)xmw, MODULEINFO->eventmask,
						  FALSE, MODULEINFO->event, moddat);
}

void
xmaid_moduledata_dup(XMaidWidget xmw, XMaidModule *module, 
					 XMaidModuleData *src, XMaidModuleData *moddat)
{
	if (MODULEINFO->dup) {
		MODULEINFO->dup(xmw, src, moddat);
		if (!ROOT && MODULEINFO->event)
			XtAddEventHandler((Widget)xmw, MODULEINFO->eventmask,
							  FALSE, MODULEINFO->event, moddat);
	} else
		xmaid_moduledata_init(xmw, module, moddat);
}



void
xmaid_module_destroy(XMaidWidget xmw, XMaidModule *module)
{
	xmaidmodule_delete(module);
}

void
xmaid_moduledata_destroy(XMaidWidget xmw, XMaidModule *module, 
						 XMaidModuleData *moddat)
{
	if (MODULEINFO->destroy)
		MODULEINFO->destroy(xmw, moddat);

	if (!ROOT && MODULEINFO->event)
		XtRemoveEventHandler((Widget)xmw, MODULEINFO->eventmask,
							 FALSE, MODULEINFO->event, moddat);

	xmaidmoduledata_delete(moddat);
}
